import sys
from .arg import *
from opensees.obj import LibCmd, Mat
import numpy as np

_patch = LibCmd("patch")


_eps = 0.00001
_huge = sys.float_info.max
_tiny = sys.float_info.min
 
def rayintersectseg(p, edge)->bool:
    """
    takes a point p=Pt() and an edge of two endpoints a,b=Pt() of a line segment returns boolean
    https://rosettacode.org/wiki/Ray-casting_algorithm#Python
    """
    a,b = edge
    if a[1] > b[1]:
        a,b = b,a
    if p[1] == a[1] or p[1] == b[1]:
        p = np.array([p[0], p[1] + _eps])
 
    intersect = False
 
    if (p[1] > b[1] or p[1] < a[1]) or (p[0] > max(a[0], b[0])):
        return False
 
    if p[0] < min(a[0], b[0]):
        intersect = True
    else:
        if abs(a[0] - b[0]) > _tiny:
            m_red = (b[1] - a[1]) / float(b[0] - a[0])
        else:
            m_red = _huge
        if abs(a[0] - p[0]) > _tiny:
            m_blue = (p[1] - a[1]) / float(p[0] - a[0])
        else:
            m_blue = _huge
        intersect = m_blue >= m_red
    return intersect


class _Polygon:
    def __init__(self, vertices):
        self.vertices = vertices

    @property
    def area(self):
        x,y = np.asarray(self.vertices).T
        return np.sum(x * np.roll(y, -1) - np.roll(x, -1) * y)/2.00

    def boundary(self):
        pass

    @property
    def centroid(self):
        x,y = np.asarray(self.vertices).T
        area = self.area
        x_cent = (np.sum((x + np.roll(x, -1)) *
                         (x*np.roll(y, -1) -
                          np.roll(x, -1)*y)))/(6.0*area)
        y_cent = (np.sum((y + np.roll(y, -1)) *
                         (x*np.roll(y, -1) -
                          np.roll(x, -1)*y)))/(6.0*area)
        return np.array((x_cent, y_cent))

    def moi(self, reference, vertices=None, area=None):
        x,y = (np.asarray(self.vertices) - np.asarray(reference)).T

        area = area or self.area
        alpha = x * np.roll(y, -1) - np.roll(x, -1) * y
        # planar moment of inertia wrt horizontal axis
        ixx = np.sum((y**2 + y * np.roll(y, -1) +
                      np.roll(y, -1)**2)*alpha)/12.00
        # planar moment of inertia wrt vertical axis
        iyy = np.sum((x**2 + x * np.roll(x, -1) +
                      np.roll(x, -1)**2)*alpha)/12.00

        ixy = np.sum((x*np.roll(y, -1)
                      + 2.0*x*y
                      + 2.0*np.roll(x, -1) * np.roll(y, -1)
                      + np.roll(x, -1) * y)*alpha)/24.
        # polar (torsional) moment of inertia
        ir = ixx + iyy
        # mass moment of inertia wrt in-plane rotation
        ir_mass = (ixx + iyy) / area

        return {'ixx': ixx, 'iyy': iyy,
                'ixy': ixy, 'ir': ir, 'ir_mass': ir_mass}
    @property
    def ixx(self):
        return self.moi(self.centroid)["ixx"]

    @property
    def iyy(self):
        return self.moi(self.centroid)["iyy"]
    
    @property
    def ixy(self):
        return self.moi(self.centroid)["ixy"]

    @property
    def iyx(self):
        return self.moi(self.centroid)["ixy"]


    def __contains__(self, p:tuple) -> bool:
        v = np.asarray(self.vertices)
        edges = np.array(
            [[v[i-1], v[i]] for i in range(len(v))]
        )
        return 1 == sum(rayintersectseg(p, edge) for edge in edges)%2


@_patch
class rect(_Polygon):
    _args = [
       Ref("matTag", field="material", 
           about="tag of previously defined material (`UniaxialMaterial`"\
                 "tag for a `FiberSection` or `NDMaterial` tag for use "\
                 "in an `NDFiberSection`)"),
       Int("numSubdivIJ", about="number of subdivisions (fibers) in the IJ direction."),
       Int("numSubdivJK", about="number of subdivisions (fibers) in the JK direction."),
       Grp("vertices", args=[
         Grp(args=[Num("yI"), Num("zI")],  about="y & z-coordinates of vertex I (local coordinate system)"),
         Grp(args=[Num("yJ"), Num("zJ")],  about="y & z-coordinates of vertex J (local coordinate system)"),
         Grp(args=[Num("yK"), Num("zK")],  about="y & z-coordinates of vertex K (local coordinate system)"),
         Grp(args=[Num("yL"), Num("zL")],  about="y & z-coordinates of vertex L (local coordinate system)"),
      ])
    ]
    def init(self):
        if len(self.vertices) == 2:
            ll, ur = self.vertices
            self.vertices = [
                #ll, [ll[0], ur[1]], ur, [ll[0], ur[1]]
                ll, [ur[0], ll[1]], ur, [ll[0], ur[1]]
            ]

@_patch
class quad(_Polygon):
    _args = [
       Ref("matTag",      field="material", 
           about="tag of previously defined material (`UniaxialMaterial` "\
                 "tag for a `FiberSection` or `NDMaterial` tag for use in an `NDFiberSection`)"),
       Grp("divs", args=[
         Int("numSubdivIJ", about="number of subdivisions (fibers) in the IJ direction."),
         Int("numSubdivJK", about="number of subdivisions (fibers) in the JK direction."),
       ]),
       Grp("vertices", args=[
         Grp(args=[Num("yI"), Num("zI")],  about="y & z-coordinates of vertex I (local coordinate system)"),
         Grp(args=[Num("yJ"), Num("zJ")],  about="y & z-coordinates of vertex J (local coordinate system)"),
         Grp(args=[Num("yK"), Num("zK")],  about="y & z-coordinates of vertex K (local coordinate system)"),
         Grp(args=[Num("yL"), Num("zL")],  about="y & z-coordinates of vertex L (local coordinate system)"),
      ])
    ]

def rhom(center, height, width, slope=None, divs=(0,0)):
    vertices = [
        [center[0] - width/2 + 1/slope*height/2, center[1] + height/2],
        [center[0] - width/2 - 1/slope*height/2, center[1] - height/2],
        [center[0] + width/2 - 1/slope*height/2, center[1] - height/2],
        [center[0] + width/2 + 1/slope*height/2, center[1] + height/2]
    ]
    return quad(vertices=vertices, div=divs)

@_patch
class circ:
    _args = [
      Ref("matTag", about="tag of previously defined material ("\
                          "`UniaxialMaterial` tag for a `FiberSection` "\
                          "or `NDMaterial` tag for use in an `NDFiberSection`)"),
      Int("numSubdivCirc", about="number of subdivisions (fibers) in "\
                                 "the circumferential direction (number of wedges)"),
      Int("numSubdivRad",  about="number of subdivisions (fibers) in the radial direction (number of rings)"),
      Grp("center",    args=[Num("yCenter"), Num("zCenter")],
          about="y & z-coordinates of the center of the circle", default=[0.0, 0.0]),
      Num("intRad",    about="internal radius", default=0.0),
      Num("extRad",    about="external radius"),
      Num("startAng",  about="starting angle", default=0.0),
      Num("endAng",    about="ending angle", default=np.pi*2),
    ]
    def __contains__(self, point):
        origin = np.asarray(self.center)
        vect = np.asarray(point) - origin
        size = np.linalg.norm(vect)
        angl = (np.arctan2(*reversed(vect))+2*np.pi) % 2*np.pi
        inside = [
            (size < self.extRad),
            (size > self.intRad),
            (angl > self.startAng),
            (angl < self.endAng),
        ]

        return all(inside)

    @property
    def Ix(self):
        """Moment of inertia"""
        return 0.25 * self.extRad**4 * np.pi
    
    @property
    def Iy(self):
        """Moment of inertia"""
        return 0.25 * self.extRad**4 * np.pi

    @property
    def J(self):
        return 0.5 * self.area * self.extRad ** 2

    @property
    def area(self):
        return np.pi*(self.extRad**2 - self.intRad**2)

    @property
    def centroid(self):
        return np.asarray(self.center)

    @property
    def ixx(self):
        return 0.25 * np.pi * (self.extRad ** 4 - self.intRad ** 4)

    @property
    def iyy(self):
        return 0.25 * np.pi * (self.extRad ** 4 - self.intRad ** 4)

    @property
    def ixy(self):
        return 0.0

layer = LibCmd("layer", {
  "circ": [
    Ref("matTag",  about="material tag of previously created material "\
                         "(UniaxialMaterial tag for a FiberSection or "\
                         "NDMaterial tag for use in an NDFiberSection)"),
    Int("numFiber",  about="number of fibers along arc"),
    Num("areaFiber", about="area of each fiber"),
    Grp("center", args=[Num("yCenter"), Num("zCenter")],
        about="y and z-coordinates of center of circular arc"),
    Num("radius", about="radius of circular arc"),
    Grp("arc", args=[
      Num("startAng",  about="starting angle (optional, default = 0.0)"),
      Num("endAng",  about="ending angle (optional, default = 360.0 - 360/$numFiber)"),
    ])
  ]
})

@layer
class line:
    _args = [
      Ref("matTag", about="""material tag of previously created material 
                (`UniaxialMaterial` tag for a `FiberSection` or `NDMaterial` 
                tag for use in an `NDFiberSection`)"""),
      Int("numFibers", about="number of fibers along line"),
      Num("areaFiber", about="area of each fiber"),
      Grp("vertices", typ=Grp, args=[
          Grp("start",args=[Num("yStart"), Num("zStart")], 
              about="""y and z-coordinates of first fiber
                       in line (local coordinate system)"""),
          Grp("end", args=[Num("yEnd"), Num("zEnd")],
              about="y and z-coordinates of last fiber in line (local coordinate system)")
      ])
    ]
    def __contains__(self, point):
        a,b = np.asarray(self.vertices)
        p = np.asarray(point)
        return np.isclose(_distance(a,p) + _distance(p,b), _distance(a,b))


def _distance(a,b):
    return np.linalg.norm(a-b)






__version__  = "0.0.1"
#
from .patch import layer

# Imports for this module
import math
import fnmatch
from .lib import Nde
from .obj import *

class Model:
    def __init__(self, ndm, ndf, nodes, elems):
        self.m_nodes = nodes
        self.m_elems = elems
        self.ndm = ndm
        self.ndf = ndf

    def get_node(self, tag):
        if isinstance(tag,(str,int)):
            return self.m_nodes[tag]

    @property
    def nodes(self):
        return self.m_nodes.values()

    @property
    def elems(self):
        return self.m_elems.values()

    @property
    def refs(self):
        return set(sub for el in self.elems for sub in el.get_refs())

    @property
    def materials(self):
        return set(a[0] for a in self.refs)

class Node:
    __slots__ = []

class Assembly:
    def __init__(self, ndm, ndf, assm={}, partials={}, units="metric", **kwds):
        self._units = units
        self.meta = kwds
        self.ndm = ndm
        self.ndf = ndf

        self.m_parameters = {}
        self.m_conns = {}
        self.m_links = {}
        self.m_nodes = {}
        self.m_bouns = {}

        if ndf == 2:
            self.prob_type = '2d-truss'
            self.dof_names: dict = { 'x': 0, 'y': 1} # Degrees of freedom
        elif ndm == 2 and ndf ==3:
            self.prob_type = '2d-frame'
            self.dof_names: dict = { 'x': 0, 'y': 1, 'xy':2}
        elif ndm == 3 and ndf ==3:
            self.prob_type = '3d-truss'
            self.dof_names: dict = { 'x': 0, 'y': 1, 'z':2}
        elif ndm == 3 and ndf ==6:
            self.prob_type = '3d-frame'
            self.dof_names: dict = { 'x': 0, 'y': 1, 'z':2, 'yz':3, 'zx':4, 'xy':5}
    

    def build(self, assm, partials):
        self.m_nodes = {
            k: Node(*v) for k,v in assm["nodes"]
        }

        self.m_conns = {
            k: partials[el.type](*el.args)
            for k, el in assm["links"]
        }

        self.m_conns.update({
            ZeroLength() for k,v in assm["bound"]
        })
    
    def get_node(self, tag):
        if isinstance(tag,(str,int)):
            return self.m_nodes[tag]

    @property
    def units(self):
        import elle.units
        return elle.units.UnitHandler(self._units)

    
    def node(self, tag, *coords, **kwds):
        self.m_nodes.update({tag: Nde(**{
            "name": tag,
            "crd": coords, 
            "boun": [0 for i in range(self.ndf)],
            **kwds,
        })})

    def conn(self, tag, typ, nodes, **kwds):
        self.m_conns.update({tag: {
            "type": typ,
            "name": tag,
            "nodes": [self.get_node(n) for n in nodes],
            **kwds
        }})

    def boun():
        pass

    def foun(self, name, typ, node):
        self.m_nodes.update({name: self.m_nodes[node](name=name,boun=[1]*self.ndf)})
        self.m_bouns.update({
            name: {
                "type": typ, 
                "nodes": [self.m_nodes[node], self.m_nodes[name]],
                "name": name,
                "boun": [1]*self.ndf
            }
        })

    def link(self, tag, typ, nodes, **kwds):
        self.m_links.update({tag: {
            "type": typ, 
            "nodes": [self.get_node(n) for n in nodes],
            **kwds
        }})

    def zero():
        pass

    def apply(self,partials,**kwds):
        elems = {}
        for con in self.m_conns.values():
            typ = con.pop("type")
            tag = con.pop("name")
            elem = partials[typ](**con)
            elem.partial = typ
            elems.update({tag: elem})

        for boun in self.m_bouns.values():
            if partials[boun["type"]] != "!fix":
                typ = boun.pop("type")
                tag = boun.pop("name")
                elem = partials[typ](**boun)
                elem.partial = typ
                elems.update({tag: elem})
            else:
                self.m_nodes.pop(boun["nodes"][-1].name)
                self.fix(boun["nodes"][0].name)

        mod = Model(self.ndm, self.ndf, self.m_nodes, elems)
        mod.dof_names = self.dof_names
        mod.partials = partials
        return mod

    def _fix_dof(self, node, dof:str):
        node.kwds["boun"][self.dof_names[dof]] = 1
        return

    def _fix_str_flags(self, node, flags):
        rxns = [
            self._fix_dof(node, dof) for dof in flags
        ]
        if len(rxns) > 1:
            return rxns
        else:
            return rxns[0]

    def _fix_int_flags(self, node, flags):
        if len(flags) != self.ndf:
            raise ValueError(f"`fix` method requires flags for all dofs ({self.ndf}) when working with ints.")
        rxns = [
            self._fix_dof(node, self.dof_nums[i]) for i in range(self.ndf)
        ]
        if len(rxns) > 1:
            return rxns
        else:
            return rxns[0]



    def fix(self, node, *dirns, x=None, y=None, z=None):
        """Define a fixed boundary condition at specified 
        degrees of freedom of the supplied node

        Parameters
        ----------
        node: anabel.Node | Sequence[anabe.Node]

        dirn: Union[Sequence[String], String]
        
        ### Example

        ```py
        # Fix all dofs at node named "abut"
        model.fix("abut")

        # Create a pinned reaction at node 2
        model.fix(2, "y")

        # Create a node and impose a roller reaction
        a = model.node("a", 0.0, 0.0)
        model.fix(a, "y", "x")

        # Fix the rotational dof in a node a 3-dof model
        model.fix("n1", 0, 0, 1)
        ```
        """
        m_nodes = self.m_nodes

        if not dirns:
            dirns = list(self.dof_names.keys())
        
        if isinstance(dirns[0], int):
            _fix =  self._fix_int_flags

        else:
            _fix = self._fix_str_flags

        if isinstance(node, str):
            nodes = [m_nodes[n] for n in fnmatch.filter(m_nodes.keys(), node)]
            nodes = filter(_filter_nodes_by_coords(x,y,z), nodes)
            return [_fix(node, dirns) for node in nodes]

        elif isinstance(node, (list,tuple)):
            return [self.fix(n, *dirns, x=x, y=y, z=z) for n in node]

        elif isinstance(node, int): 
            node = self.m_nodes[node]
            return _fix(node, dirns)
        
        else: 
            return _fix(node, dirns)

    def boun(self, node, flags: list):
        """
        Impose single-point constraints at the specified node. This 
        function is provided to give a familiar interface for users
        of platforms like FEAP and FEDEAS.

        Parameters
        ----------
        node: Union[anabel.abstract.Node, anabel.abstract.TagType]
            Node identifier

        flags: List[int]

        ### Example

        ```py
        model.boun(1, [0, 0, 1])
        ```

        """
        if isinstance(node,str):
            node = self.m_nodes[node]

        for i, dof in enumerate(self.dof_names):
            if flags[i]:
                self.fix(node, dof)



def _is_sequence(obj):
    return hasattr(type(obj), '__iter__')

def _filter_nodes_by_coords(x, y, z):
    def f(node):
        match = True
        if x is not None:
            match = match and math.isclose(node.x, x)
        if y is not None:
            match = match and math.isclose(node.y, y)
        if z is not None:
            match = match and math.isclose(node.z, z)
        return match
    return f





def _to_list(val, fmt):
    if isinstance(val, (int,float)):
        return val
    elif isinstance(val, (Int,Num)):
        return val.to_str(fmt)
    elif isinstance(val,list) or hasattr(val,"__array__"):
        #return "[" + ",".join(_to_lo
        return None

class Parameter:
    def __init__(self, name, typ):
        self.name = name
        self.typ = typ
    def __repr__(self):
        return f"${self.name}"

class Arg:
    __slots__ = ["name", "flag", "value", "field", "default", "type", "reqd"]
    def __init__(self,
        name = None,
        help = None,
        flag = None,
        reqd = True,
        type = None,
        field= None,
        about= "",
        default = None,
        **kwds
    ):
        self.name  = name
        self.flag  = [flag] if flag is not None else []
        self.value = None
        self.field = field if field is not None  else name
        self.type  = type
        self.reqd  = reqd
        self.default = default
        self.kwds = kwds
        self.about = about
        self.init()

    def __repr__(self):
        return self.__class__.__name__ + f"<{self.value}>"

    def init(self): pass

    def c_decl(self):
        pass

    def c_test_argv(self, argidx):
        return f"""(strcmp(argv[{argidx}], "{self.flag}") == 0)"""

    def c_read_argv(self, struct, argidx):
        return f"{struct}->{self.field} = argv[++{argidx}];\n"
    
    def get_value(self, parent, fmt):
        val = getattr(parent, self.field, None)
        if isinstance(val,bool):
            return str(val).lower()
        else:
            return val

    def _get_value(self, parent, value=None):
        value = self.value if value is None else value
        if value is None:
            value = self.default
        return value

    def as_tcl_list(self, value=None)->list:
        value = self._get_value(None, value)
        if isinstance(value, (type(None), )):
            if not self.reqd:
                return []
            else:
                value = f"${self.name}"
        elif isinstance(value, (Arg,)):
            if not self.reqd:
                return []
            else:
                value = f"${value.name}"
        return self.flag + [value] 

    def m_src(self, value=None):
        value = self._get_value(None,value)
        if isinstance(value,bool):
            return str(value).lower()


class Flg(Arg):
    def init(self):
        if "enum" in self.kwds:
            self.enum = self.kwds["enum"]
            if self.type is None:
                self.type = str

        self.default = False
        if self.field == self.name:
            self.field = self.name[1:]
        if "-" in self.name:
            self.flag = self.name
            self.name = self.name[1:]
    
    def c_read_argv(self, struct, argidx):
        return f"{struct}->{self.field} = true;\n"


    def as_tcl_list(self, value=None): 
        value = self.value if value is None else value
        return [self.flag] if value else []

class Num(Arg): 
    def init(self): self.type = float
    
    def c_read_argv(self, struct, argidx):
        return f"{struct}->{self.field} = argv[++{argidx}];\n"

class Int(Arg): 
    def init(self): self.type = int

class Str(Arg): 
    def init(self): self.type = str

class Chr(Arg): 
    def init(self): self.type = chr

class Ref(Arg):
    def init(self):
        if "attr" not in self.kwds: self.kwds["attr"] = "name"
    def __call__(self, *args,**kwds):
        return Ref(*args, type=self.type, attr=self.kwds["attr"], **kwds)

    def _get_value(self, parent, value=None):
        value = self.value if value is None else value
        if not isinstance(value, (str,int,float,type(None))):
            return getattr(value, self.kwds["attr"])
        else:
            return value
    
    def as_tcl_list(self, value=None): 
        value = self._get_value(None, value)
        try:
            value = getattr(value, self.kwds["attr"])
        except:
            pass
        return self.flag + [value]

class Ary(Arg): pass
class Sub(Arg): pass


class Grp(Arg):
    """Argument grouping"""
    def init(self):
        if "num" in self.kwds:
            self.args = [
                self.type(f"{self.name}{i+1}") for i in range(self.kwds["num"])
            ]
        else:
            assert isinstance(self.kwds["args"],list)
            self.args = self.kwds["args"]
        self.num = len(self.args)

    def as_tcl_list(self, value=None):
        value = self._get_value(None,value)
        if value is None:
            if self.reqd:
                value = [None]*self.num
            else:
                return []
                #value = [None]*self.num
        return self.flag + [a for arg,v in zip(self.args,value) for a in arg.as_tcl_list(v)]

class One(Arg): pass

class Blk(Grp):
    def init(self):
        self.args = []

    def as_tcl_list(self, value=None):
        value = self.value if value is None else value
        if value is None:
            value = [None]
        return self.flag + [[v.get_cmd() for v in value]]

class Tag(Int):
    def init(self):
        super().init()
        self.name = "name"
        self.field = "name"
        type = int




import skeletal

from skeletal import lib, arg
par  = arg.Parameter("Igirz", arg.Num)

con  = lib.ElasticSpring(E=20.0)
lin  = lib.LinearTransform(1)
beam = lib.ElasticBeamColumn3D(nodes=[1, 2], area=20.0, transf=lin, cMass=True)
bea2 = lib.ElasticBeamColumn3D(2, area=20.0, Iz=par, transf=2,   cMass=True, material=con)
lnk  = lib.RigidBeamLink()
abut = lib.ZeroLength3D(1, materials=[0, 2, 1], orientation=[[0.,2],[1,3]])

print(beam.get_cmd())
print(beam(cMass=False,Iz=30).get_cmd())
print(beam(Iy=40).get_cmd())

print(bea2.get_cmd())

print(abut.get_cmd())
#print(abut._args)


assm = skeletal.Assembly()

assm.node(1, 0.0, 2.0)
assm.node(2, 0.0, 5.0)
assm.node(3, 5.0, 5.0)

assm.conn(1, "girder",   (1, 2))
assm.conn(2, "abutment", (2, 3))


partials = {
    "girder":   beam,
    "abutment":  abut
}

model = assm.apply(partials)
import emit
model.ndm = 3
model.ndf = 6
print(model.refs)

print(emit.OpenSeesWriter(model).dump())





#!/bin/env python
# # Synopsis
#
# >`render.py [<options>] <model-file>
#
# >**Arpit Nema**, **Chrystal Chern**, and **Claudio Perez**
# 
#
# This script plots the geometry of a structural
# model given a SAM JSON file. The SAM JSON structure
# was developed by the NHERI SimCenter.
#
#
# # Installation
#
# The simplest way to install this script is to run
# 
#     $ python render.py --install
#
# from a terminal that has python installed, in a
# directory containing the `render.py` file. This
# will install the following packages:

REQUIREMENTS = """
pyyaml
scipy
numpy
plotly
matplotlib
"""

# ## Matlab
# In order to install the Matlab bindings, open Matlab in a
# directory containing the files `render.py` and `render.m`,
# and run the following command in the Matlab interpreter:
#
#     render --install
#
# Once this process is complete, the command `render` can be
# called from Matlab, just as described below for the command
# line.
#
# # Usage
# This script can be used either as a module, or as a command
# line utility. When invoked from the command line on
# **Windows**, {NAME} should be `python -m render`. For example:
#
#     python -m render model.json --axes 2 --view elev
#
# The script may be invoked with the following options:

NAME = "render.py"
HELP = f"""
usage: {NAME} <sam-file>
       {NAME} --install
       {NAME} [options] <sam-file>
       {NAME} [options] <sam-file> <res-file>
       {NAME} --section <py-file>#<py-object>

Generate a plot of a structural model.

Positional Arguments:
  <sam-file>                     JSON file defining the structural model.
  <res-file>                     JSON or YAML file defining a structural
                                 response.

Options:
  -s, --scale  <scale>           Set displacement scale factor.
  -d, --disp   <node>:<dof>...   Apply a unit displacement at node with tag
                                 <node> in direction <dof>.
  -V, --view   {{elev|plan|sect}}  Set camera view.
  -a, --axes   [<L><T>]<V>       Specify model axes. Only <V> is required
      --hide   <object>          Hide <object>; see '--show'.
      --show   <object>          Show <object>; accepts any of:
                                    {{origin|frames|frames.displ|nodes|nodes.displ}}

  -o, --save   <out-file>        Save plot to <out-file>.
  -c, --conf

      --install                  Install script dependencies.
      --script {{sam|res}}
  -h, --help                     Print this message and exit.



  <dof>        {{long | tran | vert | sect | elev | plan}}
               {{  0  |   1  |   2  |   3  |   4  |   5 }}
  <object>     {{origin|frames|frames.displ|nodes|nodes.displ}}
"""

EXAMPLES="""
Examples:
    Plot the structural model defined in the file `sam.json`:
        $ {NAME} sam.json

    Plot displaced structure with unit translation at nodes
    5, 3 and 2 in direction 2 at scale of 100:

        $ {NAME} -d 5:2,3:2,2:2 -s100 --axes 2 sam.json
"""

# The remainder of this script is broken into the following sections:
#
# - Data shaping / Misc.
# - Kinematics
# - Plotting
# - Command line processing
#

# Defaults
#=========
# The following configuration options are available:

Config = lambda : {
  "show_objects": ["frames", "frames.displ", "nodes"],
  "hide_objects": ["origin"],
  "sam_file":     None,
  "elem_by_type": False,
  "res_file":     None,
  "write_file":   None,
  "displ":        [],
  "scale":        100.0,
  "orientation":  [0,2,1],
  "view":         "iso",
  "plotter":      "mpl",

  "camera": {
      "view": "iso",               # iso | plan| elev[ation] | sect[ion]
      "projection": "orthographic" # perspective | orthographic
  },

  "displacements": {"scale": 100, "color": "#660505"},

  "objects": {

      "origin": {"color": "black"},

      "frames" : {
          "color": "#000000",
          "displaced": {"color": "red", "npoints": 20}
      },

      "nodes": {
          "default": {"size": 2, "color": "#000000"},
          "displaced" : {},
          "fixed"  : {},
      },
  },
  "save_options": {
      # Options for when writing to an HTML file.
      "html": {
          "include_plotlyjs": True,
          "include_mathjax" : "cdn",
          "full_html"       : True
      }
  }
}

def apply_config(conf, opts):
    for k,v in conf.items():
        if isinstance(v,dict):
            apply_conf(v, opts[k])
        else:
            opts[k] = v


# The following Tcl script can be used to create a results
# file

EIG_SCRIPT = """
for {set m 1} {$m <= 3} {incr m} {
  puts "$m:"
  foreach n [getNodeTags] {
    puts "  $n: \[[join [nodeEigenvector $n $m] {, }]\]";
  }
}
"""

import sys, os
try:
    import yaml
    import numpy as np
    Array = np.ndarray
    FLOAT = np.float32
    from scipy.linalg import block_diag
except:
    # prevent undefined variables when run in install mode
    yaml = None
    Array = list
    FLOAT =  float
NDM=3 # this script currently assumes ndm=3

# Data shaping / Misc.
#----------------------------------------------------

# The following functions are used for reshaping data
# and carrying out other miscellaneous operations.

class RenderError(Exception): pass

def wireframe(sam:dict)->dict:
    """
    Process OpenSees JSON output and return dict with the form:

        {<elem tag>: {"crd": [<coordinates>], ...}}
    """
    geom  = sam["geometry"]
    coord = np.array([n.pop("crd") for n in geom["nodes"]],dtype=FLOAT)
    nodes = {n["name"]: {**n, "crd": coord[i]} for i,n in enumerate(geom["nodes"])}
    trsfm = {t["name"]: t for t in sam["properties"]["crdTransformations"]}
    elems =  {
      e["name"]: dict(
        **e, 
        crd=np.array([nodes[n]["crd"] for n in e["nodes"]], dtype=FLOAT),
        trsfm=trsfm[e["crdTransformation"]] if "crdTransformation" in e else None
      ) for e in geom["elements"]
    }
    return dict(nodes=nodes, elems=elems, coord=coord)

def read_model(filename:str)->dict:
    import json
    with open(filename,"r") as f:
        sam = json.load(f)
    model = wireframe(sam["StructuralAnalysisModel"])
    if "RendererConfiguration" in sam:
        model["config"] = sam["RendererConfiguration"]
    return model

# Kinematics
#----------------------------------------------------

# The following functions implement various kinematic
# relations for standard frame models.

# Helper functions for extracting rotations in planes
elev_dofs = lambda v: v[[1,2]]
plan_dofs = lambda v: v[[3,4]]

def get_dof_num(dof:str, axes:list):
    try: return int(dof)
    except: return {
            "long": axes[0],
            "vert": axes[2],
            "tran": axes[1],
            "sect": axes[0]+3,
            "plan": axes[2]+3,
            "elev": axes[1]+3
    }[dof]

def elastic_curve(x: Array, v: Array, L:float)->Array:
    "compute points along Euler's elastica"
    vi, vj = v
    xi = x/L                        # local coordinates
    N1 = 1.-3.*xi**2+2.*xi**3
    N2 = L*(xi-2.*xi**2+xi**3)
    N3 = 3.*xi**2-2*xi**3
    N4 = L*(xi**3-xi**2)
    y = vi*N2+vj*N4
    return y.flatten()

def linear_deformations(u,L):
    """
    Compute local frame deformations assuming small displacements

    u: 6-vector of displacements in rotated frame
    L: element length
    """
    xi, yi, zi, si, ei, pi = range(6)    # Define variables to aid
    xj, yj, zj, sj, ej, pj = range(6,12) # reading array indices.

    elev_chord = (u[zj]-u[zi]) / L       # Chord rotations
    plan_chord = (u[yj]-u[yi]) / L
    return np.array([
        [u[xj] - u[xi]],                 # xi
        [u[ei] - elev_chord],            # vi_elev
        [u[ej] - elev_chord],            # vj_elev

        [u[pi] - plan_chord],
        [u[pj] - plan_chord],
        [u[sj] - u[si]],
    ],dtype=FLOAT)


def rotation(xyz: Array, vert=(0,0,-1))->Array:
    "Create a rotation matrix between local e and global E"
    dx = xyz[1] - xyz[0]
    L = np.linalg.norm(dx)
    e1 = dx/L
    v13 = np.atleast_1d(vert)
    v2 = -np.cross(e1,v13)
    e2 = v2 / np.linalg.norm(v2)
    v3 =  np.cross(e1,e2)
    e3 = v3 / np.linalg.norm(v3)
    return np.stack([e1,e2,e3])


def displaced_profile(
        coord: Array,
        displ: Array,        #: Displacements
        vect : Array = None, #: Element orientation vector
        glob : bool  = True, #: Transform to global coordinates
        npoints:int = 10,
    )->Array:
    n = npoints
    #          (---ndm---)
    rep = 4 if len(coord[0])==3 else 2
    Q = rotation(coord, vect)
    L = np.linalg.norm(coord[1] - coord[0])
    v = linear_deformations(block_diag(*[Q]*rep)@displ, L)
    Lnew = L+v[0,0]
    xaxis = np.linspace(0.0, Lnew, n)

    plan_curve = elastic_curve(xaxis, plan_dofs(v), Lnew)
    elev_curve = elastic_curve(xaxis, elev_dofs(v), Lnew)

    dx,dy,dz = Q@np.linspace(displ[:3], displ[6:9], n).T
    local_curve = np.stack([xaxis+dx[0], plan_curve+dy, elev_curve+dz])

    if glob:
        global_curve = Q.T@local_curve + coord[0][None,:].T

    return global_curve



# Plotting
#----------------------------------------------------

VIEWS = { # pre-defined plot views
    "plan":    dict(azim=  0, elev= 90),
    "sect":    dict(azim=  0, elev=  0),
    "elev":    dict(azim=-90, elev=  0),
    "iso":     dict(azim= 45, elev= 35)
}

def new_3d_axis():
    import matplotlib.pyplot as plt
    _, ax = plt.subplots(1, 1, subplot_kw={"projection": "3d"})
    ax.set_autoscale_on(True)
    ax.set_axis_off()
    return ax

def add_origin(ax,scale):
    xyz = np.zeros((3,3))
    uvw = np.eye(3)*scale
    ax.quiver(*xyz, *uvw, arrow_length_ratio=0.1, color="black")
    return ax

def set_axis_limits(ax):
    "Find and set axes limits"
    aspect = [ub - lb for lb, ub in (getattr(ax, f'get_{a}lim')() for a in 'xyz')]
    aspect = [max(a,max(aspect)/8) for a in aspect]
    ax.set_box_aspect(aspect)

def plot_skeletal(frame, axes=None):
    if axes is None: axes = [0,2,1]
    props = {"frame": {"color": "grey", "alpha": 0.6}}
    ax = new_3d_axis()
    for e in frame["elems"].values():
        x,y,z = e["crd"].T[axes]
        ax.plot(x,y,z, **props["frame"])
    return ax


def plot_nodes(frame, displ=None, axes=None, ax=None):
    if axes is None: axes = [0,2,1]
    ax = ax or new_3d_axis()
    displ = displ or {}
    Zero = np.zeros(NDM)
    props = {"color": "black",
             "marker": "s",
             "s": 2,
             "zorder": 2}

    coord = frame["coord"]
    for i,n in enumerate(frame["nodes"].values()):
        coord[i,:] += displ.get(n["name"],Zero)[:3]

    x,y,z = coord.T[axes]
    ax.scatter(x, y, z, **props)
    return ax

def plot_displ(frame:dict, res:dict, ax=None, axes=None):
    props = {"color": "#660505"}
    ax = ax or new_3d_axis()
    if axes is None: axes = [0,2,1]
    for el in frame["elems"].values():
        # exclude zero-length elements
        if "zero" not in el["type"].lower():
            glob_displ = [
                u for n in el["nodes"] 
                #   extract displ from node, default to ndf zeros
                    for u in res.get(n,[0.0]*frame["nodes"][n]["ndf"])
            ]
            vect = el["trsfm"]["vecInLocXZPlane"]
            x,y,z = displaced_profile(el["crd"], glob_displ, vect=vect)[axes]
            ax.plot(x,y,z, **props)
    return ax

class Plotter:
    def __init__(self,model, opts, axes=None):
        self.model = model
        if axes is None: axes = [0,2,1]
        self.axes = axes
        self.opts = opts

class MplPlotter(Plotter):
    def __init__(self, **kwds):
        pass
    def get_section_layers(self, section):
        import matplotlib.collections
        import matplotlib.lines as lines
        collection = []
        for layer in section.layers:
            if hasattr(layer, "plot_opts"):
                options = layer.plot_opts
            else:
                options = dict(linestyle="--", color="k")
            collection.append(
                lines.Line2D(*np.asarray(layer.vertices).T, **options))
        return collection

    def get_section_patches(self, section):
        import matplotlib.collections
        import matplotlib.patches as mplp
        collection = []
        for patch in section.patches:
            name = patch.__class__.__name__.lower()
            if "circ" in name:
                if patch.intRad:
                    width = patch.extRad - patch.intRad
                    collection.append(mplp.Annulus(patch.center, patch.extRad, width))
                else:
                    collection.append(mplp.Circle(patch.center, patch.extRad))
            else:
                collection.append(mplp.Polygon(patch.vertices))
        return matplotlib.collections.PatchCollection(
            collection,
            facecolor="grey",
            edgecolor="grey",
            alpha=0.3
        )

    def plot_section(self,
        section,
        show_properties=False,
        plain=True,
        show_quad=True,
        show_dims=True,
        annotate=True,
        ax = None,
        fig = None,
        **kwds
    ):
        """Plot a composite cross section."""    
        import matplotlib.pyplot as plt
        if plain:
            show_properties = show_quad = show_dims = False

        if show_properties:
            fig = plt.figure(constrained_layout=True)
            gs = fig.add_gridspec(1,5)
            axp = fig.add_subplot(gs[0,3:-1])
            label = "\n".join(["${}$:\t\t{:0.4}".format(k,v) for k,v in self.properties().items()])
            axp.annotate(label, (0.1, 0.5), xycoords='axes fraction', va='center')
            axp.set_autoscale_on(True)
            axp.axis("off")

            ax = fig.add_subplot(gs[0,:3])
        else:
            fig, ax = plt.subplots()

        if ax is None:
            fig, ax = plt.subplots()
        ax.set_autoscale_on(True)
        ax.set_aspect(1)

        #x_max = 1.01 * max(v[0] for p in self.patches for v in p.vertices if hasattr(p,"vertices"))
        #y_max = 1.05 * max(v[1] for p in self.patches for v in p.vertices if hasattr(p,"vertices"))
        #y_min = 1.05 * min(v[1] for p in self.patches for v in p.vertices if hasattr(p,"vertices"))

        #ax.set_xlim(-x_max, x_max)
        #ax.set_ylim( y_min, y_max)
        ax.axis("off")
        # add shapes
        ax.add_collection(self.get_section_patches(section,**kwds))
        for l in self.get_section_layers(section):
            ax.add_line(l)
        # show centroid
        #ax.scatter(*section.centroid)
        # show origin
        ax.scatter(0, 0)
        plt.show()
        
        return fig, ax

class GnuPlotter(Plotter):
    def plot_frames(self):
        file=sys.stdout
        print("""
        set term wxt
        unset border
        unset xtics
        unset ytics
        unset ztics
        set view equal xyz
        splot "-" using 1:2:3 with lines
        """,file=file)
        coords = self._get_frames()
        np.savetxt(file, coords)

    def _get_frames(self):
        axes = self.axes
        model = self.model
        props = {"color": "#808080", "alpha": 0.6}
        coords = np.zeros((len(model["elems"])*3,NDM))
        coords.fill(np.nan)
        for i,e in enumerate(model["elems"].values()):
            coords[3*i:3*i+2,:] = e["crd"][:,axes]
        return coords



class PlotlyPlotter(Plotter):
    def plot(x,y,**opts):
        pass

    def plot_frames():
        pass

    def _get_displ(self,res:dict):
        frame = self.model
        axes = self.axes
        props = {"color": "red"}
        N = 10
        coords = np.zeros((len(frame["elems"])*(N+1),NDM))
        coords.fill(np.nan)
        for i,el in enumerate(frame["elems"].values()):
            # exclude zero-length elements
            if "zero" not in el["type"].lower():
                glob_displ = [
                    u for n in el["nodes"]
                    #   extract displ from node, default to ndf zeros
                        for u in res.get(n,[0.0]*frame["nodes"][n]["ndf"])
                ]
                vect = el["trsfm"]["vecInLocXZPlane"]
                coords[(N+1)*i:(N+1)*i+N,:] = displaced_profile(el["crd"], glob_displ, vect=vect, npoints=N)[axes].T
        x,y,z = coords.T
        return {
            "type": "scatter3d",
            "mode": "lines",
            "x": x, "y": y, "z": z,
            "line": {"color":props["color"]},
            "hoverinfo":"skip"
        }
    def make_hover_data(self, data, ln=None):
        if ln is None:
            items = np.array([d.values for d in data])
            keys = data[0].keys()
        else:
            items = np.array([list(data.values())]*ln)
            keys = data.keys()
        return {
            "hovertemplate": "<br>".join(f"{k}: %{{customdata[{v}]}}" for v,k in enumerate(keys)),
            "customdata": list(items),
        }

    def _get_nodes(self):
        x,y,z = self.model["coord"].T[self.axes]
        keys  = ["tag",]
        nodes = np.array(list(self.model["nodes"].keys()),dtype=FLOAT)[:,None]
        return {
                "name": "nodes",
                "x": x, "y": y, "z": z,
                "type": "scatter3d","mode": "markers",
                "hovertemplate": "<br>".join(f"{k}: %{{customdata[{v}]}}" for v,k in enumerate(keys)),
                "customdata": list(nodes),
                "marker": {
                    "symbol": "square",
                    **self.opts["objects"]["nodes"]["default"]
                }
        }

    def _get_frame_labels(self):
        coords = self._frame_coords.reshape(-1,4,3)[:,-3]
        x,y,z = coords.T
        keys  = ["tag",]
        frames = np.array(list(self.model["elems"].keys()),dtype=FLOAT)[:,None]
        return {
                "name": "frames",
                "x": x, "y": y, "z": z, 
                "type": "scatter3d","mode": "markers",
                "hovertemplate": "<br>".join(f"{k}: %{{customdata[{v}]}}" for v,k in enumerate(keys)),
                "customdata": frames,
                "opacity": 0
                #"marker": {"opacity": 0.0,"size": 0.0, "line": {"width": 0.0}}
        }

    def _get_frames(self, elems=None):
        N = 4
        axes = self.axes
        elems = elems or self.model["elems"]
        props = {"color": "#808080", "alpha": 0.6}
        coords = np.zeros((len(elems)*N,NDM))
        coords.fill(np.nan)
        for i,e in enumerate(elems.values()):
            coords[N*i:N*i+N-1,:] = np.linspace(*e["crd"][:,axes], N-1)
        self._frame_coords = coords
        x,y,z = coords.T
        return [{
            "type": "scatter3d",
            "mode": "lines",
            "x": x, "y": y, "z": z,
            "line": {"color":props["color"]},
            "hoverinfo":"skip"
        }]


def plot_plotly(model, axes=None, displ=None, opts={}):
    import plotly.graph_objects as go
    plt = PlotlyPlotter(model,axes=axes,opts=opts)
    if opts["elems_by_type"]:
        frames = plt._get_frames()
    else:
        frames = plt._get_frames()
    labels = plt._get_frame_labels()
    nodes = plt._get_nodes()
    fig = go.Figure(dict(
            #go.Scatter3d(**plot_skeletal_plotly(model,axes)),
            data=[*frames, labels, nodes] + ([plt._get_displ(displ)] if displ else []),
            layout=go.Layout(
                scene=dict(aspectmode='data',
                     xaxis_visible=False,
                     yaxis_visible=False,
                     zaxis_visible=False,
                     camera=dict(
                         projection={"type": opts["camera"]["projection"]}
                     )
                ),
                showlegend=False
            )
        ))
    return fig

# Script functions
#----------------------------------------------------

# Argument parsing is implemented manually because in
# the past I have found the standard library module
# `argparse` to be slow.

def parse_args(argv)->dict:
    opts = Config()
    if os.path.exists(".render.yaml"):
        with open(".render.yaml", "r") as f:
            presets = yaml.load(f, Loader=yaml.Loader)

        apply_config(presets,opts)

    args = iter(argv[1:])
    for arg in args:
        try:
            if arg == "--help" or arg == "-h":
                print(HELP)
                sys.exit()

            elif arg == "--gnu":
                opts["plotter"] = "gnu"

            elif arg == "--install":
                try: install_me(next(args))
                # if no directory is provided, use default
                except StopIteration: install_me()
                sys.exit()

            elif arg[:2] == "-d":
                node_dof = arg[2:] if len(arg) > 2 else next(args)
                for nd in node_dof.split(","):
                    node, dof = nd.split(":")
                    opts["displ"].append((int(node), get_dof_num(dof, opts["orientation"])))

            elif arg[:2] == "-s":
                opts["scale"] = float(arg[2:]) if len(arg) > 2 else float(next(args))

            elif arg == "--scale":
                opts["scale"] = float(next(args))

            elif arg == "--axes":
                vert = int(next(args))
                tran = 2 if vert == 1 else 1
                opts["orientation"][1:] = [tran, vert]

            elif arg == "--show":
                opts["show_objects"].extend(next(args).split(","))

            elif arg == "--hide":
                opts["show_objects"].pop(next(args))
            
            elif arg[:2] == "-V":
                opts["view"] = arg[2:] if len(arg) > 2 else next(args)
            elif arg == "--view":
                opts["view"] = next(args)

            elif arg == "--elem-by-type":
                opts["elem_by_type"] = True

            #elif arg[:2] == "-m":
            #    opts["mode"] = int(arg[2]) if len(arg) > 2 else int(next(args))

            elif arg[:2] == "-o":
                opts["write_file"] = arg[2:] if len(arg) > 2 else next(args)

            #elif arg == "--displ-only":
            #    opts["displ_only"] = True

            # Final check on options
            elif arg[0] == "-":
                raise RenderError(f"ERROR - unknown option '{arg}'")

            elif not opts["sam_file"]:
                opts["sam_file"] = arg

            else:
                opts["res_file"] = arg

        except StopIteration:
            # `next(args)` was called without successive arg
            raise RenderError(f"ERROR -- Argument '{arg}' expected value")
    return opts

def install_me(install_opt=None):
    import os
    import subprocess
    if install_opt == "dependencies":
        subprocess.check_call([
            sys.executable, "-m", "pip", "install", *REQUIREMENTS.strip().split("\n")
        ])
        sys.exit()
    try:
        from setuptools import setup
    except ImportError:
        from distutils.core import setup

    sys.argv = sys.argv[:1] + ["develop", "--user"]
    print(sys.argv)

    setup(name="render",
          version="0.0.1",
          description="",
          long_description=HELP,
          author="", author_email="", url="",
          py_modules=["render"],
          scripts=["render.py"],
          license="",
          install_requires=[*REQUIREMENTS.strip().split("\n")],
    )

TESTS = [
    (False,"{NAME} sam.json -d 2:plan -s"),
    (True, "{NAME} sam.json -d 2:plan -s50"),
    (True, "{NAME} sam.json -d 2:3    -s50"),
    (True, "{NAME} sam.json -d 5:2,3:2,2:2 -s100 --axes 2 sam.json")
]


def render(sam_file, res_file=None, **opts):
    if sam_file is None:
        raise RenderError("ERROR -- expected positional argument <sam-file>")

    model = read_model(sam_file)

    if "config" in model:
        apply_config(model["config"], opts)
    
    axes = opts["orientation"]

    if opts["plotter"] == "gnu":
        GnuPlotter(model, axes).plot_frames()
        sys.exit()

    #for obj in opts["show_objects"]:
    #    plt[obj]

    ax = plot_skeletal(model,axes=axes)
       
    if res_file is not None:
        from urllib.parse import urlparse
        res_path = urlparse(res_file)
        with open(res_path[2], "r") as f:
            res = yaml.load(f,Loader=yaml.Loader)
        if res_path[4]: # query parameters passed
            res = res[int(res_path[4].split("=")[-1])]

    else:
        res = {}
    for n,d in opts["displ"]:
        v = res.setdefault(n,[0.0]*model["nodes"][n]["ndf"])
        if d < 3: # translational dof
            v[d] += 1.0
        else:
            v[d] += 0.1

    # apply scale
    scale = opts["scale"]
    if scale != 1.0:
        for n in res.values():
            for i in range(len(n)):
                n[i] *= scale


    ax = plot_nodes(model, res, ax=ax, axes=axes)
    if res:
        plot_displ(model, res, axes=axes, ax=ax)

    # Handle plot formatting
    set_axis_limits(ax)
    ax.view_init(**VIEWS[opts["view"]])
    if "origin" in opts["show_objects"]: add_origin(ax, scale)

    if opts["write_file"]:
    # write plot to file if file name provided
        if "html" in opts["write_file"]:
            fig = plot_plotly(model,axes,displ=res,opts=opts)
            import plotly
            print(str(id(model)),file=sys.stderr)
            html = plotly.io.to_html(fig, div_id=str(id(model)), **opts["save_options"]["html"])
            with open(opts["write_file"],"w+") as f:
                f.write(html)
        else:
            ax.figure.savefig(opts["write_file"])
    else:
    # otherwise show in new window
        import matplotlib.pyplot as plt
        plt.show()
    return ax

# Main script
#----------------------------------------------------
# The following code is only executed when the file
# is invoked as a script.

if __name__ == "__main__":
    opts = parse_args(sys.argv)
    if "json" in opts["sam_file"]:
        # Plot structural model
        try:
            render(**opts)
        except (FileNotFoundError,RenderError) as e:
            print(e, file=sys.stderr)
            print(f"         Run '{NAME} --help' for more information", file=sys.stderr)
            sys.exit()
    else:
        # Plot a cross section
        from urllib.parse import urlparse
        file = urlparse(opts["sam_file"])
        with open(file.path, "r") as f:
            script = f.read()
        scope = {}
        exec(script, scope)
        plt = MplPlotter()
        plt.plot_section(scope[file.fragment])







from .opensees import OpenSeesWriter
from .writer import ModelWriter
from .fedeas import FEDEAS_Writer

class JSON(ModelWriter):
    def dump(self):
        import json

        out = {
            "StructuralAnalysisModel":{
                "properties": {"crdTransformations": [
                    
                ]},
                "geometry": {
                    "nodes": [
                        {**n, "crd": [float(i) for i in n.crd]} for n in self.model.nodes
                    ],
                    "elements": [
                    ]
                }
            }
        }
        return json.dumps(out, indent=2)



# Claudio Perez
from .arg import *
from .obj import *

Dof = Int

class Yng(Num):
    """Young's elastic modulus."""
    def init(self):
        super().init()
        self.name = "E"
        self.field = "elastic_modulus"

class Area(Num):
    def init(self):
        super().init()
        self.name = "A"
        self.field = "area"


Nde = cmd("Node", "node", [
    Tag(),
    Grp("crd", type=Num, args=[Num("x"),Num("y"), Num("z")]),
    Grp("mass", flag="-mass", reqd=False, default=[0.0]*6, args=[
        Num("x"),Num("y"), Num("z"), Num("x"),Num("y"), Num("z")
    ]),
])

ElasticSpring = Uni("ElasticUniaxialMaterial",
    "Elastic",
    args = [
        Tag(),
        Yng(),
        Num("eta",  field="damp_tangent",     reqd=False, default=0.0, about="damping tangent"),
        Num("Eneg", field="negative_modulus", reqd=False),
        Num("G",    field="shear_modulus",    reqd=False),
    ]
)

Steel02 = Uni("Steel02",
    "Steel02",
    args = [
      Tag(),
      Num("Fy", about="yield strength"),
      Num("E0", about="initial elastic tangent"),
      Num("b", about="strain-hardening ratio (ratio between post-yield tangent and initial elastic tangent"),

        # params (list (float)) parameters to control the transition from 
        # elastic to plastic branches. params=[R0,cR1,cR2].
        # Recommended values: R0=between 10 and 20, cR1=0.925, cR2=0.15"),

      Grp("a", about="isotropic hardening parameters", args=[
        Num("a1", about="""
                  increase of compression yield envelope as proportion
                  of yield strength after a plastic strain of `a2∗(Fy/E0)`"""
        ),
        Num("a2", about="see explanation under `a1`."),
        Num("a3", about="""
                  increase of tension yield envelope as proportion
                  of yield strength after a plastic strain of `a4∗(Fy/E0)`"""
        ),
        Num("a4", about="see explanation under `a3`."),
      ]),
    ],
    about="""
    This command is used to construct a uniaxial 
    Giuffre-Menegotto-Pinto steel material object 
    with isotropic strain hardening."""
)


Concrete02 = Uni("Concrete02",
    "Concrete02",
    args=[
      Tag(about="integer tag identifying material"),
      Num("fpc",   about="concrete compressive strength at 28 days (compression is negative)"),
      Num("epsc0", about="concrete strain at maximum strength"),
      Num("fpcu",  about="concrete crushing strength"),
      Num("epsU",  about="concrete strain at crushing strength"),
      Num("lamda", about="ratio between unloading slope at `epscu` and initial slope"),
      Num("ft",    about="tensile strength"),
      Num("Ets",   about="tension softening stiffness (absolute value) " +
                       "(slope of the linear tension softening branch)"),
    ],
    about="""This command is used to construct a uniaxial Kent-Scott-Park
        concrete material object with degraded linear unloading/reloading
        stiffness according to the work of Karsan-Jirsa and no tensile
        strength. (REF: Fedeas)."""
    )

ZeroLength3D = Ele("ZeroLength3D",
    "zeroLength",
    args = [
        Tag(),
        Grp("nodes", args=[
            Ref("iNode", type=Nde,  attr="name", about=""),
            Ref("jNode", type=Nde,  attr="name", about=""),
        ]),
        Grp("materials", flag="-mat", type=Ref(type=Uni, attr="name"), num=6),
        Grp("dofs",      flag="-dir", type=Dof, num=6, default=[
              "$dx", "$dy", "$dz", "$rx", "$ry", "$rz"
        ]),
        Grp("orientation", reqd=False, flag="-orient", args=[
            Grp("x",  type=Num, num=3),
            Grp("yp", type=Num, num=3)
        ]),
        Int("do_rayleigh", flag="-doRayleigh", default=0)
    ],
    refs=["materials"]
)

ElasticBeamColumn3D = Ele("ElasticBeamColumn3D",
    "elasticBeamColumn",
    args = [
        Tag(),
        Grp("nodes", args=[
          Ref("iNode", type=Nde,  attr="name", about=""),
          Ref("jNode", type=Nde,  attr="name", about=""),
        ]),
        Area(alt="section"),
        Yng( alt="material"),
        Num("G",    field="shear_modulus",   about="", alt="material"),
        Num("J",    field="torsion_modulus", about="", alt="section"),
        #Grp("moi", ctype="struct", args=[
          Num("Iy",                about="", alt="section"),
          Num("Iz", field="Ix",    about="", alt="section"),
        #]),
        Ref("geom",  field="transform",    type=Trf, attr="name"),
        Num("mass_density", flag="-mass", default=0.0, reqd=False, about="element mass per unit length"),
        Flg("-cMass", field="consistent_mass",
            about="Flag indicating whether to use consistent mass matrix.")
    ],
    refs=["transform"],
    alts=[
        Ref("material", type=Mat),
        Ref("section",  type=Sec)
    ]
)

LinearTransform = Trf("LinearTransform",
  "Linear",
  args = [
    Tag("name", "Tag used to identify the transform object."),
    Grp("vecxz", type=Num, num=3),
    Grp("joint_offsets", reqd=False, flag="-jntOffset", type=Grp, args=[
        Grp(type=Num,num=3, default=[0.0]*3),
        Grp(type=Num,num=3, default=[0.0]*3)
    ])
  ],
)

RigidBeamLink = Lnk("RigidBeamLink",
    "beam", [Tag(), Grp("nodes", type=Ref, num=2)]
)


#FiberSection = Sec("FiberSection",
#    "fiber",
#    args = [
#        Tag(),
#        Num("GJ", flag="-GJ", field="torsional_stiffness", optional=True, 
#            about="linear-elastic torsional stiffness assigned to the section (optional, default = no torsional stiffness)"),
#        Blk("fibers", type=Cmd, defn=dict(
#           fiber=LibCmd("fiber"),
#           layer=LibCmd("layer")
#          )
#        )
#    ]
#)

redirect = Cmd("redirect",[
      Blk("commands"),
      One(optional=True, enum=[
         Str("filename", flag=">"),
         Str("filename", flag=">>"),
         Grp("shell_pipe", flag="|", args=[
            Str("shell_name", optional=True),
            Blk("shell_block", flag="|"),
         ]),
         Str("variable")
      ])
    ])

# region = Cmd("region", about="""
#     The region command is used to label a group of nodes and elements. This command
#     is also used to assign rayleigh damping parameters to the nodes and elements in
#     this region. The region is specified by either elements or nodes, not both. If
#     elements are defined, the region includes these elements and the all connected
#     nodes, unless the `-eleOnly` option is used in which case only elements are
#     included. If nodes are specified, the region includes these nodes and all
#     elements of which all nodes are prescribed to be in the region, unless the
#     `-nodeOnly` option is used in which case only the nodes are included. 
#     """,
#     #region $regTag <-ele ($ele1 $ele2 ...)> 
#     #               <-eleOnly ($ele1 $ele2 ...)> 
#     #               <-eleRange $startEle $endEle> 
#     #               <-eleOnlyRange $startEle $endEle> 
#     #               <-node ($node1 $node2...)> 
#     #               <-nodeOnly ($node1 $node2 ...)> 
#     #               <-nodeRange $startNode $endNode>
#     #               <-nodeOnlyRange $startNode $endNode> 
#     #               <-node all> 
#     #               <-rayleigh $alphaM $betaK $betaKinit $betaKcomm>
# 
#     # region 1 -ele 1 5 -eleRange 10 15
#     # region 2 -node 2 4 6 -nodeRange 9 12 
#     args=[
#     Tag(),
#     Grp(flag="-ele",
#         about="tags of selected elements in domain to be included in region (optional, default: omitted)"),
#     Grp(flag="-eleRange",
#         args=[Ref("startEle"), Ref("endEle")],
#         about="tag for start and end elements -- range of selected elements in domain"),
#     Grp(flag="-node",
#         optional=True,
#         about="tags of selected nodes in domain to be included in region (optional, default: omitted)"),
#     Grp(flag="-nodeRange",
#         optional=True,
#         args=[Ref("startNode"), Ref("endNode")], 
#         about="tag for start and end nodes -- range of nodes in domain"),
#     Grp(flag="-rayleigh", 
#         optional=True,
#         args=["alphaM", "betaK", "betaKinit", "betaKcomm"],
#         about="Arguments to define Rayleigh damping matrix (optional, default: zero)"),
#     ])





class ModelWriter:
    def __init__(self, model):
        if not hasattr(model, "header"):
            model.header = ""

        self.model = model
    
    def dump(self, definitions={}, **kwds):
        model = self.model
        c = self.comment_char
        ndm, ndf = model.ndm, model.ndf
        try:
            cmds  = model.header.replace("\n", f"\n{c} ") + "\n"
        except:
            cmds = ""


        head = "\n" + self.heading(1, "Initializations")
        body = self.dump_initialize(definitions=definitions, **kwds)
        if body and not body.isspace():
            cmds += "\n" + head + body
         
        head  = self.heading(1, "Materials")
        body  = self.dump_materials(**kwds)
        if body and not body.isspace():
            cmds += "\n" + head + body

        head  = "\n" + self.heading(1, "Elements")
        body  = self.dump_elements(**kwds)
        if body and not body.isspace():
            cmds += "\n" + head + body

        cmds += "\n" + self.heading(1, "Assemblage")
        cmds += self.dump_connectivity(**kwds)

        cmds += "\n" + self.heading(1, "Constraints")
        cmds += self.dump_constraints(**kwds)

        return "\n\n".join([cmds])

    def dump_materials(self, definitions={})->str:
        return ""

    def dump_sections(self, definitions={})->str:
        raise Exception("This model is not yet implemented by the dispatched writer")

    def dump_elements(self, *elems)->str:
        return ""

    def dump_assembly(self)->str:
        pass

    def heading(self, level, text):
        c = self.comment_char
        if level == 1:
            return f"""{c}{'='*60}\n{c} {text}\n{c}{'='*60}\n"""
        elif level == 2:
            return f"\n{c} {text}\n{c}{'-'*50}"





#!/bin/env python
import numpy as np
import matplotlib.pyplot as plt

def show(): plt.show()

def section(s, **kwds):
    return MplPlotter().plot_section(s, **kwds)

class MplPlotter:
    def __init__(self, **kwds):
        pass
    def get_section_layers(self, section, **kwds):
        import matplotlib.collections
        import matplotlib.lines as lines
        collection = []
        for layer in section.layers:
            if hasattr(layer, "plot_opts"):
                options = layer.plot_opts
            else:
                options = dict(linestyle="--", color="k", **kwds)
            collection.append(
                lines.Line2D(*np.asarray(layer.vertices).T, **options))
        return collection

    def get_section_patches(self, section, **kwds):
        import matplotlib.collections
        import matplotlib.patches as mplp
        collection = []
        for patch in section.patches:
            name = patch.__class__.__name__.lower()
            if "circ" in name:
                if patch.intRad:
                    width = patch.extRad - patch.intRad
                    collection.append(mplp.Annulus(patch.center, patch.extRad, width))
                else:
                    collection.append(mplp.Circle(patch.center, patch.extRad))
            elif "rect" in name:
                vertices = np.asarray(patch.vertices)
                collection.append(mplp.Rectangle(
                    vertices[0], *(vertices[2]-vertices[0])
                ))
            else:
                collection.append(mplp.Polygon(np.asarray(patch.vertices)))
        return matplotlib.collections.PatchCollection(
            collection,
            facecolor="grey",
            edgecolor="grey",
            alpha=0.3
        )

    def plot_section(self,
        section,
        show_properties=False,
        plain=True,
        show_quad=True,
        show_dims=True,
        annotate=True,
        ax = None,
        fig = None,
        **kwds
    ):
        """Plot a composite cross section."""    
        if plain:
            show_properties = show_quad = show_dims = False

        if show_properties:
            fig = plt.figure(constrained_layout=True)
            gs = fig.add_gridspec(1,5)
            axp = fig.add_subplot(gs[0,3:-1])
            label = "\n".join([
                "${}$:\t\t{:0.4}".format(k,v) for k,v in self.properties().items()
            ])
            axp.annotate(label, (0.1, 0.5), xycoords='axes fraction', va='center')
            axp.set_autoscale_on(True)
            axp.axis("off")

            ax = fig.add_subplot(gs[0,:3])
        else:
            fig, ax = plt.subplots()

        if ax is None:
            fig, ax = plt.subplots()
        ax.set_autoscale_on(True)
        ax.set_aspect(1)

        #x_max = 1.01 * max(v[0] for p in section.patches for v in p.vertices if hasattr(p,"vertices"))
        #y_max = 1.05 * max(v[1] for p in section.patches for v in p.vertices if hasattr(p,"vertices"))
        #y_min = 1.05 * min(v[1] for p in section.patches for v in p.vertices if hasattr(p,"vertices"))

        #ax.set_xlim(-x_max, x_max)
        #ax.set_ylim( y_min, y_max)
        ax.axis("off")
        # add shapes
        ax.add_collection(self.get_section_patches(section))
        for l in self.get_section_layers(section, **kwds):
            ax.add_line(l)
        # show centroid
        #ax.scatter(*section.centroid)
        # show origin
        ax.scatter(0, 0)
        #plt.show()
        
        return ax






from opensees.obj import LibCmd, Cmd
from opensees.arg import *

_section = LibCmd("section")

class FiberCollection:
    def __init__(self, fibers):
        self.fibers = fibers
    def __contains__(self, point):
        return any(point in fiber for fiber in self.fibers)

@_section
class FiberSection(FiberCollection):
    _args = [
        Tag(),
        Num("GJ", flag="-GJ", field="torsional_stiffness", optional=True, 
            about="linear-elastic torsional stiffness assigned to the section (optional, default = no torsional stiffness)"),
        Blk("fibers", default=[], type=Cmd, defn=dict(
           fiber=LibCmd("fiber"),
           layer=LibCmd("layer")
          )
        )
    ]

    def add_patch(self, patch):
        self.fibers.append(patch)

    def add_patches(self, patch):
        self.fibers.extend(patch)


    @property
    def patches(self):
        return [p for p in self.fibers if p.get_cmd()[0] == "patch"]

    @property
    def layers(self):
        return [p for p in self.fibers if p.get_cmd()[0] == "layer"]

    @property
    def area(self):
        return sum(i.area for i in self.patches)
    
    @property
    def centroid(self):
        return sum(i.centroid * i.area for i in self.patches) / self.area
    
    @property
    def ixy(self):
        return sum(
            p.ixy + p.centroid[0]*p.centroid[1]*p.area for p in self.patches
        )
    
    @property
    def iyy(self):
        return sum(
            p.iyy + p.centroid[0]**2*p.area for p in self.patches
        )

    @property
    def ixx(self):
        return sum(
            p.ixx + p.centroid[1]**2*p.area for p in self.patches
        )

    @property
    def moi(self):
        return [
            [p.moi[i] + p.centroid[i]**2*p.area for i in range(2)] + [p.moi[-1]]
            for p in self.patches
        ]







from .writer import ModelWriter

class OpenSeesWriter(ModelWriter):
    def __init__(self, model=None):
        self.model = model
        self.comment_char = "#"

    def dump_initialize(self, definitions={}):
        cmds = "# Parameters\n" + "\n".join(f"set {k} {v};" for k,v in definitions.items()) + "\n\n"
        ndm, ndf = self.model.ndm, self.model.ndf
        dof_keys = "dx dy dz rx ry rz"
        dofs = " ".join( str(i) for i in range(1, ndf + 1))
        cmds += f"# Create ModelBuilder (with {ndm} dimensions and {ndf} DOF/node)\n"
        cmds += f"model BasicBuilder -ndm {ndm} -ndf {ndf}\n"
        cmds += f"lassign {{{dofs}}} {dof_keys}"
        return cmds

    #@classmethod 
    def dump_elements(self, *elems, definitions={}):
        transforms = set()
        cmds = "\n".join(f"set {k} {v};" for k,v in definitions.items()) + "\n"
        for i, el in enumerate(elems):
            #el.init()
            if el.name is None:
                el.name = i+1
            try:
                cmds += "\nelement " + " ".join(str(p) for p in el.ops_elem.serialize(tag=i))
            except:
                cmds += "\n" + " ".join(el.get_cmd_str()) + "\n"

            if hasattr(el, "_transform") and el._transform:
                transforms.update({el._transform})

        return "".join(" ".join(t.get_cmd_str()) + "\n" for t in transforms) + cmds

    @classmethod 
    def dump_sections(self, *sections, definitions={}):
        cmds = "\n".join(f"set {k} {v};" for k,v in definitions.items()) + "\n"
        for sect in sections:
            cmds += " ".join(sect.get_cmd_str()) + "\n"
        return cmds
    
    def dump_materials(self, *materials, definitions={}):
        cmds = "\n".join(f"set {k} {v};" for k,v in definitions.items()) + "\n"
        all_materials = self.model.materials
        for i,mat in enumerate(all_materials):
            if mat.name is None:
                mat.name = i+1
            try:
                cmds += " ".join(mat.get_cmd_str()) + "\n"
            except:
                try:
                    cmds += "\nuniaxialMaterial " + " ".join(str(p) for p in mat.serialize(tag=i))
                except:
                    pass
        return cmds

    def dump_constraints(self, definitions={}):
        cmds = "\n".join(f"set {k} {v};" for k,v in definitions.items()) + "\n"
        for n in self.model.nodes:
            if 'boun' in n.kwds and any(n.kwds["boun"]):
                cmds += f"fix {n.name} {' '.join(str(i) for i in n.kwds['boun'])}\n"

        #model = self.model
        #for n in model.nodes:
        #    if any(n.constraints):
        #        cmds += f"fix {n.tag} {' '.join(str(i) for i in n.constraints)}\n"

        #if hasattr(model, "links"):
        #    for link in model.links:
        #        cmds += f"rigidLink {link['type']} {' '.join(str(n.tag) for n in link['nodes'])}\n"
        return cmds

    def dump_connectivity(self, definitions={}):
        cmds = "\n".join(f"set {k} {v};" for k,v in definitions.items()) + "\n"
        for i,nd in enumerate(self.model.nodes):
            nd.name = i
            cmds +=  " ".join(nd.get_cmd_str()) + "\n"
        cmds += self.dump_elements(*self.model.elems)
        return cmds





import numpy as np
# Kinematics
#----------------------------------------------------
shapes = {
    "vert": [
        (lambda xi, L: 1.-3.*xi**2+2.*xi**3),
        (lambda xi, L: 3.*xi**2-2*xi**3)
    ],
    "tran": [
        (lambda xi, L: 1.-3.*xi**2+2.*xi**3),
        (lambda xi, L: 3.*xi**2-2*xi**3)
    ],
    "sect": [
        (lambda xi, L: 1-xi),
        (lambda xi, L: xi)
    ],
    "plan": [
        (lambda xi, L: L*(xi-2.*xi**2+xi**3)),
        (lambda xi, L: L*(xi**3-xi**2))
    ],
    "elev": [
        (lambda xi, L: L*(xi-2.*xi**2+xi**3)),
        (lambda xi, L: L*(xi**3-xi**2))
    ],
}
# The following functions implement various kinematic
# relations for standard frame models.

# Helper functions for extracting rotations in planes
elev_dofs = lambda u: u[[1,2]]
plan_dofs = lambda u: u[[3,4]]

def get_dof_num(dof:str, axes:list):
    try: return int(dof)
    except: return {
            "long": axes[0],
            "vert": axes[2],
            "tran": axes[1],
            "sect": axes[0]+3,
            "plan": axes[2]+3,
            "elev": axes[1]+3
    }[dof]

def elastic_curve(x: Array, v: Array, L:float)->Array:
    "compute points along Euler's elastica"
    vi, vj = v
    xi = x/L                        # local coordinates
    N1 = 1.-3.*xi**2+2.*xi**3
    N2 = L*(xi-2.*xi**2+xi**3)
    N3 = 3.*xi**2-2*xi**3
    N4 = L*(xi**3-xi**2)
    y = vi*N2+vj*N4
    return y.flatten()

def linear_deformations(u,L):
    """
    Compute local frame deformations assuming small displacements

    u: 6-vector of displacements in rotated frame
    L: element length
    """
    xi, yi, zi, si, ei, pi = range(6)    # Define variables to aid
    xj, yj, zj, sj, ej, pj = range(6,12) # reading array indices.

    elev_chord = (u[zj]-u[zi]) / L       # Chord rotations
    plan_chord = (u[yj]-u[yi]) / L
    return np.array([
        [u[xj] - u[xi]],                 # xi
        [u[ei] - elev_chord],            # vi_elev
        [u[ej] - elev_chord],            # vj_elev

        [u[pi] - plan_chord],
        [u[pj] - plan_chord],
        [u[sj] - u[si]],
    ],dtype=FLOAT)


def rotation(xyz: Array, vert=(0,0,-1))->Array:
    "Create a rotation matrix between local e and global E"
    dx = xyz[1] - xyz[0]
    L = np.linalg.norm(dx)
    e1 = dx/L
    v13 = np.atleast_1d(vert)
    v2 = -np.cross(e1,v13)
    e2 = v2 / np.linalg.norm(v2)
    v3 =  np.cross(e1,e2)
    e3 = v3 / np.linalg.norm(v3)
    return np.stack([e1,e2,e3])


def displaced_profile(
        coord: Array,
        displ: Array,        #: Displacements
        vect : Array = None, #: Element orientation vector
        glob : bool  = True, #: Transform to global coordinates
        npoints:int = 10,
    )->Array:
    n = npoints
    #          (---ndm---)
    rep = 4 if len(coord[0])==3 else 2
    Q = rotation(coord, vect)
    L = np.linalg.norm(coord[1] - coord[0])
    v = linear_deformations(block_diag(*[Q]*rep)@displ, L)
    Lnew = L+v[0,0]
    xaxis = np.linspace(0.0, Lnew, n)

    plan_curve = elastic_curve(xaxis, plan_dofs(v), Lnew)
    elev_curve = elastic_curve(xaxis, elev_dofs(v), Lnew)

    #dy,dz = Q[1:,1:]@np.linspace(displ[1:3], displ[7:9], n).T
    dx,dy,dz = Q@np.linspace(displ[:3], displ[6:9], n).T
    local_curve = np.stack([xaxis+dx[0], plan_curve+dy, elev_curve+dz])

    if glob:
        global_curve = Q.T@local_curve + coord[0][None,:].T

    return global_curve




from .writer import ModelWriter
from skeletal.arg import *
from skeletal.obj import Component, LibCmd, Ele, walk_refs

Frm = LibCmd("ElemData",
    args = [
        Flg("geom", enum=("linear")),
        # Vec("yornt")
    ],
    rels = {
        "ElasticBeamColumn3D": {
            ".transform.vecxz": {".yorn": lambda v: None},
            ".transform.name":   ".geom",
            ".transform.joint_offset":   ".JntOff",
            ".mass_density":     ".rho"
        }
    }
)

dict(
    LE3dFrm = Frm("LE3dFrm",
        "_",
        args = [],
    )
)

def fmt_val(arg,value):
    if isinstance(value, list):
        if isinstance(value[0],list):
            return "[" + \
                    "; ".join(", ".join(str(v) for v in l) for l in value) \
                    + "]"
        else:
            return "[" + \
                    ", ".join(str(v) for v in value) \
                    + "]"
    else:
        return str(value)

#import anabel.elements

import numpy as np
from datetime import datetime

FEDEASmap = {
    "2D beam": "Lin2dFrm",
    #anabel.elements.ElasticBeam: "LEFrame",
    #anabel.elements.ZeroLength: "zeroLength",
    "2D truss": "LETruss",
}

class FEDEAS_Writer(ModelWriter):
    def __init__(self, model, filename=None, simple=False):
        self.time_stamp = datetime.now().strftime("%Y%m%d-%H%M%S")
        self.model = model
        self.filename = filename
        self.simple = simple

        self.comment_char = "%"

        self.boun = []

    def dump_initialize(self, definitions={}):
        Domain = self.model
        # script += 'function [Model,ElemData,Loading] = {}()'.format(self.filename)
        script = ""
        #script += "\n" + "CleanStart\n"
        script += "\n".join(f"{k} = {v};" for k, v in definitions.items()) + "\n"
        return script

        # Node Definitions

    def dump_connectivity(self):
        Domain = self.model
        script = "% Node Definitions\n"
        for i, node in enumerate(Domain.nodes):
            node.name = i+1
            script += "XYZ({},:) = [{}];\n".format(
                i + 1, " ".join(f"{x:8.8}" for x in node.crd)
            )
            if any(node.kwds["boun"]):
                self.boun.append((node, node.kwds["boun"]))

        # Connectivity
        script += "\n% Connections\n"
        for i, elem in enumerate(Domain.elems):
            #ni = Domain.nodes.index(elem.nodes[0]) + 1
            #nj = Domain.nodes.index(elem.nodes[1]) + 1
            ni  = elem.nodes[0].name
            nj  = elem.nodes[1].name
            script += "CON({},:) = [{:8} {:8}];\n".format(i + 1, ni, nj)

        return script

    def dump_constraints(self):
        # Fixities
        script = ""
        Domain = self.model
        script += "%                " + " ".join(f"{i:4}" for i in Domain.dof_names.keys()) + "\n"
        for node, boun in self.boun:
            rx = " ".join(f"{i:4}" for i in boun)
            script += "BOUN({:3},:) = [{}];\n".format(node.name, rx)


        script += "\n" + "\n% Create model"
        if self.simple:
            script += "\n" + "Model = Create_SimpleModel(XYZ, CON, BOUN, ElemName);"
        else:
            script += "\n" + "Model = Create_Model(XYZ, CON, BOUN, ElemName);"
        return script

    def dump_materials(self,**kwds):
        script,sub_script = "", ""
        typs = set()
        ident = "    "
        for value in self.model.materials:
            script += "Components.{}('{}') = struct(...\n".format(
                value._cmd[0], value.name, ident
            )
            if value._cmd[0] not in typs:
                sub_script += "Components.{} = containers.Map('KeyType', 'char', 'ValueType', 'any');\n".format(
                    value._cmd[0]
                )
                typs = typs.union({value._cmd[0]})
            for sub_arg in value._args:

                sub_val = getattr(value,sub_arg.field)
                if sub_val is not None:
                    val_str = fmt_val(sub_arg, sub_val)
                    script += "{}'{}', {},...\n".format(
                         ident, sub_arg.name, val_str
                    )
            script = script[:script.rfind(",...")]
            script += ");\n"
        return sub_script + script
    

    def dump_elements(self):
        # Element types
        Domain = self.model
        from collections import defaultdict
        partials = defaultdict(list)
        for i, elem in enumerate(Domain.elems):
            elem.name = i+1
            partials[elem.partial].append(elem.name)

        script = "% Specify element types\n"
        elem_script = ""
        script += f"\nElemData = cell({len(Domain.elems)},1);\n"
        for partial_name, elem in Domain.partials.items():
            if elem == "!fix":
                continue
            elem_typ = elem.type if hasattr(elem, "type") else elem.__class__.__name__
            elem_dic = {}
            for arg in elem._args:
                field = elem_dic.get(arg,arg.field)
                value = arg.get_value(elem,".m")
                if value is None:
                    pass
                elif isinstance(arg, Ref):
                    script += "ElemType.{}.{} = Components.{}('{}');\n".format(
                        partial_name, arg.name, value._cmd[0], value.name
                    )
                elif isinstance(arg, Grp) and Grp.type == Ref:
                    pass

                else:
                    script += "ElemType.{}.{} = {};\n".format(partial_name, arg.name, value)
            elem_script += f"for el={partials[partial_name]!r}, ElemName{{el}}='{elem_typ}'; ElemData{{el}} = ElemType.{partial_name}; end;\n"
            script += "\n"

        return (script+elem_script).replace("$", "")

    def dump_loading(self):
        # Element Loads
        Domain = self.model
        script = "\n" + "\n%% Element loads"
        # Nodal Loads
        script += "\n" + "\n%% Nodal loads"
        script += "\n" + "Pf = zeros({});".format(Domain.nf)
        for node in Domain.nodes:
            p = node.p_vector()
            for i, dof in enumerate(node.dofs):
                if p[i] != 0.0:
                    script += "\n" + "Pf({}) = {};".format(dof, p[i])
        script += "\n" + "Loading.Pref = Pf;"
        return script

    def write(self):
        f = open(self.filename, "w+")
        f.write(self.string())
        f.close()




from  .arg  import *
from typing import get_type_hints

class Component:
    @staticmethod
    def get_value(cls, arg, value):
        pass

    def get_cmd(self):
        args = [
            a for arg in self._args
                for a in arg.as_tcl_list(value=getattr(self, arg.field))
        ]
        return self._cmd + args

    def get_cmd_str(self):
        for i in self.get_cmd():
            if isinstance(i,list):
                yield "{\n"
                for j in i:
                    yield f"{j}"
                yield "\n}\n"
            else:
                yield str(i)

    def __call__(self, **kwds):
        return self.copy(**kwds)

    def copy(self, **kwds):
        partial = self.__class__()
        for field in self.__slots__:
            if field in kwds:
                setattr(partial, field, kwds.pop(field))
            else:
                setattr(partial, field, getattr(self, field))
        partial.kwds.update(kwds)
        return partial

    def _init(self):
        for arg in self._args:
            field = arg.field
            if getattr(self, field) is None:
                try: setattr(self, field, self.kwds[arg.name])
                except Exception as e: pass
            if getattr(self, field) is None:
                try: setattr(self, field, arg._get_value(None))
                except Exception as e: pass
            if getattr(self, field) is None:
                try: setattr(self, field, getattr(
                     getattr(self, arg.kwds["alt"]), field))
                except Exception as e: pass

    def get_refs(self):
        for arg in self._args:
            if arg.field in self._refs:
                if isinstance(arg, Ref):
                    yield (getattr(self, arg.field), arg.type)
                elif isinstance(arg, Grp):
                    typ = arg.type.type
                    for i in getattr(self, arg.field):
                        yield (i, typ)

class Cmd:
    cmd = None
    def __init__(self, cmd , defs=None):
        self.cmd = cmd


def cmd(cls, cmd, args, refs=[], **ops):
    fields = [arg.field for arg in args] 
    alts = {arg.kwds["alt"] for arg in args if "alt" in arg.kwds}
    obj = struct(cls, fields, args, alts, refs=refs)
    obj._cmd  = [cmd]
    return obj


class LibCmd(Cmd):
    cmd = None
    def __init__(self, cmd , subs={}, args=None, rels=None, defs=None):
        self.cmd = cmd
        self.args = [] if args is None else args
        self.rels = [] if rels is None else rels

        for sub in subs:
            setattr(self, sub, self(sub.title(), sub, args=subs[sub]))

    def __call__(self, cls, name=None, args=None, refs=[], **opts):
        if isinstance(cls,str):
            typ = []
        else:
            args = cls._args
            typ = [cls]
            cls = name = cls.__name__

        args = self.args + args
        fields = [arg.field for arg in args]
        if "alts" in opts: fields += [a.field for a in opts["alts"]]
        alts = {arg.kwds["alt"] for arg in args if "alt" in arg.kwds}
        obj = struct(cls, fields, args, alts, refs=refs, parents=typ)
        obj._cmd  = [self.cmd, name]
        obj.kwds = opts
        setattr(self, name, obj)
        return obj


def struct(name, fields, args = None, alts=None, refs=[], parents=[]):
    import textwrap
    template = textwrap.dedent("""\
    class {name}({parents}):
        __slots__ = {fields!r}
        def __init__(self, {fields_none}, **kwds):
            {self_fields} = {args}
            self.kwds = kwds
            self._refs = {refs!r}
            self._init()
            if hasattr(self,"init"): self.init()

        def keys(self): return self.__slots__
        def __getitem__(self, idx):
            return getattr(self, idx)

    """).format(
        name=name,
        refs=refs,
        fields=fields,
        parents = ",".join(["Component"]+[p.__name__ for p in parents]),
        args=','.join(fields),
        fields_none=','.join(f"{f}=None" for f in fields),
        self_fields=','.join('self.' + f for f in fields)
    )
    d = {'fields': fields, "Arg": Arg, "Component": Component}
    d.update({c.__name__: c for c in parents})
    exec(template, d)
    d[name]._args = args
    return d[name]

Mat = LibCmd("nDMaterial")
Uni = LibCmd("uniaxialMaterial")
Ele = LibCmd("element")
Trf = LibCmd("geomTransf")
Sec = LibCmd("section")
Lnk = LibCmd("rigidLink")

def walk_refs(parent):
    P, R = "", ""
    for ref in parent.get_refs():
        print(ref)
        p, r = walk_refs(ref[0])
        P = p + P
        R = r + R
    return P, R

